/*
 * Copyright (c) 2017, Red Hat Inc. All rights reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package org.openjdk.gcbench.retain;

import org.openjdk.jmh.annotations.*;

import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.TimeUnit;

@Warmup(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Measurement(iterations = 5, time = 1, timeUnit = TimeUnit.SECONDS)
@Fork(1)
@BenchmarkMode(Mode.AverageTime)
@OutputTimeUnit(TimeUnit.NANOSECONDS)
@Threads(Threads.MAX)
@State(Scope.Benchmark)
public class Tree {

    Object retain;

    @Param({"100000"})
    private int size;

    @Setup
    public void setup() {
        List<Node> nodes = new ArrayList<>();
        for (int c = 0; c < size; c++) {
            nodes.add(new Node(null, null));
        }

        while (nodes.size() > 1) {
            List<Node> newNodes = new ArrayList<>();
            for (int c = 0; c < nodes.size(); c += 2) {
                if (c + 1 < nodes.size()) {
                    newNodes.add(new Node(nodes.get(c), nodes.get(c + 1)));
                } else {
                    // non-balanced tree
                    newNodes.add(new Node(nodes.get(c), null));
                }
            }
            nodes = newNodes;
        }

        retain = nodes.get(0);
    }

    @Benchmark
    public Object test() {
        // allocation pressure to trigger GCs
        return new Object();
    }

    private static class Node {
        final Node left;
        final Node right;
        final Object payload;
        public Node(Node left, Node right) {
            this.left = left;
            this.right = right;
            this.payload = new Object();
        }
    }

}
